import clingo, clingo.ast
import clingo_ast_util
import pr_settings, pr_restriction
from pr_stdio import dprint, dinput

class Transformer:
    """This class is the basis of rewrites on the logic program
    
    The class stores a reference of every rule in the program.
    It also redirects rules to the ProjectionTransformer for rewriting.
    """
    def __init__(self, builder, predictor):
        self.builder = builder
        self.predictor = predictor
        self.built = False
        self.stms = []
        self.projection_cache = {}
        self.used_predicate_symbols = set()
        self.total_projection_count = 0
        self.projection_counts = {}
        self.pending_projection_counts = {}
        self.projected_rule_id_count = 0
        self.body_functions = {}
        self.head_functions = {}
        self.no_rewrite = False
        if pr_settings.args.oracle:
            self.answer_queue = pr_settings.answer_queue[:] # Clone list
            self.queue_done = False
            self.tried_y = False

    def visit(self, x):
        if isinstance(x, clingo.ast.AST):
            if x.type == clingo.ast.ASTType.Rule:
                return self.handle_rule(x)
            else:
                return x
        elif x is None:
            return x
        else:
            raise TypeError("unexpected type")

    def handle_rule(self, rule):
        if self.no_rewrite:
            return rule
        if pr_settings.args.restrict == 'alpha':
            tau = pr_restriction.alpha
        elif pr_settings.args.restrict == 'beta':
            tau = pr_restriction.beta
        else:
            raise Exception('Restriction "%s" is not supported' % pr_settings.args.restrict)
        rule, tau_rules = self.projection(rule, tau)
        for tau_rule in tau_rules:
            self.add_stm(tau_rule)
        return rule

    def prepare_stm(self, stm):
        if pr_settings.args.no_rewrite:
            parsed_stm = stm
            if pr_settings.args.oracle:
                pr_settings.oracle_done = True
        else:
            parsed_stm = self.visit(stm)
        return self.add_stm(parsed_stm)
    
    def add_stm(self, stm):
        self.stms.append(stm)
    
    def build_program(self):
        if self.built:
            raise Exception('Program has already been built for this transformer')
        # self.make_projected_functions_uniq() FIXME Not perfect with linking, need to look into
        for stm in self.stms:
            self.builder.add(stm)
        self.built = True
    
    def equal_projection_rules(self, ruleA, ruleB):
        """Used to check equality between projection so the same projection is not performed twice"""
        if not isinstance(ruleA, clingo.ast.AST) or not isinstance(ruleB, clingo.ast.AST):
            return False
        if not ruleA.type == clingo.ast.ASTType.Rule or not ruleB.type == clingo.ast.ASTType.Rule:
            return False
        # Both are at least rules
        if not hasattr(ruleA, "proj_metadata") or not hasattr(ruleB, "proj_metadata"):
            return False
    
    def make_projected_functions_uniq(self):
        """Makes all functions located in self.projected_functions have unique names within the program"""
        dprint('~~~MAKING PROJECTED FUNCTIONS UNIQUE~~~', 3)
        for projected_rule_id, head_fun in self.head_functions.items():
            dprint('Checking [%s]: %s' % (projected_rule_id, head_fun), 3)
            if projected_rule_id not in self.body_functions or len(self.body_functions[projected_rule_id]) == 0:
                continue
            fun_name = self.body_functions[projected_rule_id][0].name # We just need to grab any of the names, the first one works
            while fun_name in self.used_predicate_symbols: # Find unique name
                fun_name += "'"
            self.used_predicate_symbols.add(fun_name)
            if fun_name != self.body_functions[projected_rule_id][0].name:
                dprint('New name: %s' % fun_name, 3)
                dprint('Updating body functions: %s' % self.body_functions[projected_rule_id], 4)
            for body_fun in self.body_functions[projected_rule_id]: # Rename all body functions to the unique name
                body_fun.name = fun_name
            head_fun.name = fun_name # Rename head function in the projected rule to the unique name
        dprint('', 3)
    
    def add_to_cache(self, rule):
        """Adds the rule to the cache
        Returns True if the rule was added to the cache, False otherwise
        """
        wrapped_rule = clingo_ast_util.ASTWrapper(rule)
        if wrapped_rule in self.projection_cache:
            # Already did this projection somewhere
            projected_rule_id = self.projection_cache[wrapped_rule]
            # dprint('Reusing rule: %s [%d]' % (rule, projected_rule_id))
            return (False, projected_rule_id)
        else:
            # New projection
            projected_rule_id = self.projected_rule_id_count
            self.projected_rule_id_count += 1
            self.projection_cache[wrapped_rule] = projected_rule_id
            # dprint('New rule: %s [%d]' % (rule, projected_rule_id))
            return (True, projected_rule_id)

    def add_head_and_body_function_link(self, projected_rule_id, head_fun, body_fun):
        """Adds a function to be associated with the projected rule id to ensure any function renaming will apply"""
        if projected_rule_id not in self.head_functions:
            self.head_functions[projected_rule_id] = head_fun
        self.add_body_function_link(projected_rule_id, body_fun)
        setattr(head_fun, 'projected_rule_id', projected_rule_id)

    def add_body_function_link(self, projected_rule_id, body_fun):
        """ 
        The constructor for ComplexNumber class. 
  
        Parameters: 
           real (int): The real part of complex number. 
           imag (int): The imaginary part of complex number.    
        """
        if projected_rule_id not in self.body_functions:
            self.body_functions[projected_rule_id] = []
        self.body_functions[projected_rule_id].append(body_fun)
        setattr(body_fun, 'projected_rule_id', projected_rule_id)

    def expand_selected_var(self, selected_var, tau, alpha_dict, projection_target, context_rule=None):
        """
        Expands a selected var to a set of vars corresponding to all vars that can be projected alongside it

        Parameters:
            selected_var (clingo.ast.AST): The selected var to expand upon
            alpha_dict (dict): The dictionary containing the cached alpha information
            projection_target (clingo.ast.AST): The AST target to project from
            context_rule (clingo.ast.AST): Operations to be performed with respect to the context_rule, if provided (optional)
        """
        projectable_vars = set([selected_var])
        for wrapped_var, wrapped_literal_set in alpha_dict.items():
            if wrapped_var == selected_var:
                continue
            if pr_restriction.alpha(self, projection_target, wrapped_var.ast).issubset(tau(self, projection_target, selected_var.ast, context_rule=context_rule)):
                dprint('Including %s in the projection' % wrapped_var, 2)
                projectable_vars.add(wrapped_var)
        dprint('Expansion: %s -> %s' % (selected_var, projectable_vars), 2)
        return projectable_vars

    def projection(self, projection_target, tau, context_rule=None, ignore_vars=None):
        """
        Performs projection on a rule. If a context_rule is provided, tau-literals will be calculated with respect to it.

        Parameters:
            rule (clingo.ast.AST): The candidate rule to perform projection on.
            tau (function): The restriction operator to apply to projection.
            aggregate_target (clingo.ast.AST): The aggregate target to project (optional).
            ignore_vars (set): The set of variables to ignore.

        Returns:
            tuple: rule, tau_rules
        """
        alpha_dict_builder = pr_restriction.AlphaDictionaryBuilder(self, ignore_aggregates=False)
        alpha_dict = alpha_dict_builder.get_alpha_dict(projection_target)
        unsafe_vars = alpha_dict_builder.variable_counter.unsafe_vars
        if type(projection_target) != list: # FIXME Don't repeat like this?
            alpha_dict_builder = pr_restriction.AlphaDictionaryBuilder(self, ignore_aggregates=True)
            alpha_dict = alpha_dict_builder.get_alpha_dict(projection_target)
        aggregate_ignore_vars = alpha_dict_builder.variable_counter.get_aggregate_ignore_vars()
        if pr_settings.args.aggregates:
            AggregateTransformer(self, tau, unsafe_vars.union(aggregate_ignore_vars)).aggregate_projection(projection_target)
        if ignore_vars is not None:
            for ignore_var in ignore_vars:
                if ignore_var in alpha_dict:
                    del alpha_dict[ignore_var]
        tau_rules = []
        while len(alpha_dict) > 0:
            if self.total_projection_count == pr_settings.args.debug_after:
                pr_settings.args.debug = 1
            if pr_settings.args.limit <= self.total_projection_count:
                return projection_target, []
            dprint('Alpha-Dict: %s' % alpha_dict, 2)
            if pr_settings.args.heuristic == 'variable':
                #(selected_var, selected_literal_set) = self.get_largest_variable_expansion(tau, alpha_dict, projection_target, context_rule)
                (selected_var, selected_literal_set) = self.get_smallest_variable_expansion(tau, alpha_dict, projection_target, context_rule)
            elif pr_settings.args.heuristic == 'ascending':
                (selected_var, selected_literal_set) = self.get_smallest_alpha_pair(alpha_dict)
            else:
                (selected_var, selected_literal_set) = self.get_largest_alpha_pair(alpha_dict)
            dprint('Selected: (%s, %s)' % (selected_var, selected_literal_set), 2)
            
            projectable_vars = self.expand_selected_var(selected_var, tau, alpha_dict, projection_target, context_rule)
            dprint('SELECTED EXPANSION: %s -> %s' % (selected_var, projectable_vars))
            
            base_proj_size = len(selected_literal_set) # NOTE: Does not include bounding literals!
            self.queue_projection_count(base_proj_size)
            dprint('Projecting %s' % projection_target, 2)
            target_before = str(projection_target)
            replacement_target, tau_rule, pn_tau_rule = self.project(projection_target, projectable_vars, tau, context_rule=context_rule)
            
            equivalent_variables = False
            non_conservative_proj = False
            if tau_rule is not None:
                if hasattr(replacement_target, 'type') and replacement_target.type == clingo.ast.ASTType.Rule:
                    if clingo_ast_util.find_variables(pn_tau_rule.body) == clingo_ast_util.find_variables(projection_target.body):
                        equivalent_variables = True 
                elif isinstance(replacement_target, list):
                    if clingo_ast_util.find_variables(pn_tau_rule.body) == clingo_ast_util.find_variables(projection_target):
                        equivalent_variables = True
                else:
                    raise Exception('Unable to analyze type: %s (%s)' % (type(replacement_target), replacement_target))
                if pr_settings.args.conservative and len(clingo_ast_util.find_variables(pn_tau_rule.head).intersection(clingo_ast_util.find_variables(replacement_target.body))) > 0:
                    dprint('Conservative situation found', 2)
                    non_conservative_proj = True
            if equivalent_variables or non_conservative_proj:
                if equivalent_variables:
                    dprint('Skipping projection: tau rule contains same variables as base target', 2)
                else:
                    dprint('Skipping projection: non-conservative projection found with conservative mode enabled', 2)
                self.clear_pending_counts()
                self.projection_cache.pop(clingo_ast_util.ASTWrapper(pn_tau_rule), None)
                del alpha_dict[selected_var]
                continue

            for projectable_var in projectable_vars:
                del alpha_dict[projectable_var]
            
            if self.predictor is not None:
                before_pred = self.predictor.predict_rule(projection_target)
                proj_preds = self.predictor.predict_all_rules([tau_rule, replacement_target])
                tau_pred = proj_preds[0]
                replacement_pred = proj_preds[1]
            else:
                before_pred = -1
                tau_pred = -1
                replacement_pred = -1

            if tau_rule is not None:
                dprint('\nAdding rule: \033[93m%s\033[0m (size=%s)' % (tau_rule, tau_pred))
            else:
                dprint('')
            dprint('Before projection: \033[94m%s\033[0m (size=%s)' % (target_before, before_pred))
            dprint('After  projection: \033[92m%s\033[0m (size=%s)\n' % (replacement_target, replacement_pred))
            if pr_settings.args.random_rules:
                prompt_answer = pr_settings.rule_random.choice(['y','n'])
                pr_settings.answer_queue.append(prompt_answer)
            elif pr_settings.args.oracle:
                global oracle_done
                if len(self.answer_queue) > 0:
                    prompt_answer = self.answer_queue.pop(0)
                    dprint('Answering with %s (%d)' % (prompt_answer, len(self.answer_queue)))
                    if len(self.answer_queue) <= 0:
                        oracle_done = True # Start by assuming we've done all projections
                else:
                    if not self.tried_y:
                        dprint('Trying "y"...')
                        prompt_answer = 'y'
                        self.tried_y = True
                    else:
                        dprint('Queue done...')
                        prompt_answer = 'n'
                        oracle_done = False # There are still more projections to do
            else:
                if not self.predictor or pr_settings.args.debug:
                    prompt_answer = dinput('Would you like to perform this projection? [y/n] ') # 'y' if not debug
                else:
                    if tau_pred + replacement_pred < before_pred:
                        prompt_answer = 'y'
                    else:
                        prompt_answer = 'n'
                if self.predictor or pr_settings.args.debug:
                    # Only store if in debug / using Predictor
                    pr_settings.answer_queue.append(prompt_answer)
            if prompt_answer == 'y*':
                dprint('Projecting this rule & stopping debug mode (future projections will occur)')
                pr_settings.args.debug = 0
            elif prompt_answer == 'n':
                # NOTE WARNING: Selecting to skip will eliminate ALL candidate variables from consideration, possible need to rethink/revise to make skipping better
                dprint('Skipping...')
                self.clear_pending_counts()
                self.projection_cache.pop(clingo_ast_util.ASTWrapper(pn_tau_rule), None)
                continue
            projection_target = replacement_target
            if tau_rule is not None:
                tau_rules.append(tau_rule)
            
            self.commit_projection_counts()
        return projection_target, tau_rules
    
    def project(self, projection_target, projectable_vars, tau, context_rule=None):
        replacement_target = clingo_ast_util.copy(projection_target, self)
        alpha_literals = pr_restriction.alpha(self, replacement_target, projectable_vars)
        if len(alpha_literals) == 0:
            raise Exception('Tried to project %s when no matching alpha literals were found' % projectable_vars)
        if tau == pr_restriction.alpha:
            tau_literals = alpha_literals
        else:
            tau_literals = tau(self, replacement_target, projectable_vars, context_rule=context_rule)
        loc = next(iter(alpha_literals)).ast.location # TODO Probably make this something like "N/A"
        tau_rule_body = [clingo_ast_util.copy(literal_wrapper.ast) for literal_wrapper in tau_literals]
        alpha_vars = clingo_ast_util.find_variables(alpha_literals)
        remaining_vars = alpha_vars - projectable_vars
        tau_function = clingo.ast.Function(loc, 'p_fun', [var_wrapper.ast for var_wrapper in remaining_vars], False)
        tau_rule = clingo.ast.Rule(loc,\
                        clingo.ast.Literal(loc, clingo.ast.Sign.NoSign, clingo.ast.SymbolicAtom(tau_function)),\
                        tau_rule_body)
        body_tau_function = clingo_ast_util.copy(tau_function)

        pn_tau_rule = clingo_ast_util.copy(tau_rule)
        # Normalize
        if not pr_settings.args.no_normalize:
            tau_rule = clingo_ast_util.RuleNormalizer().normalize(tau_rule) # TODO What if the literals are in different orders?
        # prenamed_tau_rule = clingo_ast_util.copy(tau_rule) # Copy so that renaming the head predicate doesn't apply to this rule
        is_rule_added, projected_rule_id = self.add_to_cache(pn_tau_rule)
        tau_function.name = 'p_fun_%d' % projected_rule_id
        body_tau_function.name = tau_function.name
        if is_rule_added:
            # It's a new rule
            dprint('New rule [%d]: %s' % (projected_rule_id, tau_rule), 3)
            dprint('(Projecting out %s)' % projectable_vars, 3)
        else:
            # Existing rule, don't add
            dprint('Existing rule [%d]: %s' % (projected_rule_id, tau_rule), 3)
            tau_rule = None

        if clingo_ast_util.is_rule(replacement_target):
            if pr_settings.args.old_mode:
                replacement_target.body = [body_atom for body_atom in replacement_target.body if clingo_ast_util.ASTWrapper(body_atom) not in alpha_literals]
            else:
                replacement_target.body = [body_atom for body_atom in replacement_target.body if clingo_ast_util.ASTWrapper(body_atom) not in tau_literals]
            replacement_list = replacement_target.body
        elif isinstance(replacement_target, list):
            if pr_settings.args.old_mode:
                replacement_target = [body_atom for body_atom in replacement_target if clingo_ast_util.ASTWrapper(body_atom) not in alpha_literals]
            else:
                replacement_target = [body_atom for body_atom in replacement_target if clingo_ast_util.ASTWrapper(body_atom) not in tau_literals]
            replacement_list = replacement_target
        else:
            raise Exception('Unable to replace type: %s (%s)' % (type(replacement_target), replacement_target))
        replacement_list.append(clingo.ast.Literal(loc, clingo.ast.Sign.NoSign, clingo.ast.SymbolicAtom(body_tau_function)))
        return replacement_target, tau_rule, pn_tau_rule

    def queue_projection_count(self, size):
        curr_count = self.pending_projection_counts.get(size, 0)
        self.pending_projection_counts[size] = curr_count + 1
    
    def commit_projection_counts(self):
        for size, p_count in self.pending_projection_counts.items():
            curr_count = self.projection_counts.get(size, 0)
            self.projection_counts[size] = curr_count + p_count
            self.total_projection_count += p_count
        self.clear_pending_counts()
    
    def clear_pending_counts(self):
        self.pending_projection_counts = {}

    def get_largest_alpha_pair(self, alpha_dict):
        """Returns the largest alpha-set"""
        smallest_size = -1
        smallest_pairs = []
        for var, literal_set in alpha_dict.items():
            check_size = len(literal_set)
            if check_size > smallest_size:
                smallest_size = check_size
                smallest_pairs = [(var, literal_set)]
            elif pr_settings.args.random and check_size == smallest_size:
                smallest_pairs.append((var, literal_set))
        return pr_settings.heuristic_random.choice(smallest_pairs)
    
    def get_smallest_alpha_pair(self, alpha_dict):
        """Returns the smallest alpha-set"""
        smallest_size = float('inf')
        smallest_pairs = []
        for var, literal_set in alpha_dict.items():
            check_size = len(literal_set)
            if check_size < smallest_size:
                smallest_size = check_size
                smallest_pairs = [(var, literal_set)]
            elif pr_settings.args.random and check_size == smallest_size:
                smallest_pairs.append((var, literal_set))
        return pr_settings.heuristic_random.choice(smallest_pairs)
    
    def get_largest_variable_expansion(self, tau, alpha_dict, projection_target, context_rule=None): # TODO Cache expansions or something...
        largest_size = -1
        largest_pairs = []
        candidate_vars = alpha_dict.items()
        for var, literal_set in candidate_vars:
            expanded_vars = self.expand_selected_var(var, tau, alpha_dict, projection_target, context_rule)
            size = len(expanded_vars)
            if size > largest_size:
                largest_size = size
                largest_pairs = [(var, literal_set)]
            elif pr_settings.args.random and size == largest_size:
                largest_pairs.append((var, literal_set))
        return pr_settings.heuristic_random.choice(largest_pairs)

    def get_smallest_variable_expansion(self, tau, alpha_dict, projection_target, context_rule=None): # TODO Cache expansions or something...
        smallest_size = float('inf')
        smallest_pairs = []
        candidate_vars = alpha_dict.items()
        for var, literal_set in candidate_vars:
            expanded_vars = self.expand_selected_var(var, tau, alpha_dict, projection_target, context_rule)
            size = len(expanded_vars)
            dprint('%s: %s' % (var, expanded_vars))
            if size < smallest_size:
                smallest_size = size
                smallest_pairs = [(var, literal_set)]
            elif pr_settings.args.random and size == smallest_size:
                smallest_pairs.append((var, literal_set))
        return pr_settings.heuristic_random.choice(smallest_pairs)



class AggregateTransformer(clingo_ast_util.ASTVisitor):

    def __init__(self, base_transformer, tau, unsafe_vars=None):
        self.base_transformer = base_transformer
        self.tau = tau
        self.unsafe_vars = unsafe_vars
    
    def aggregate_projection(self, rule):
        self.rule = rule
        return self.visit(rule)

    def visit_BodyAggregate(self, body_aggregate):
        for aggregate_element in body_aggregate.elements:
            ignore_vars = clingo_ast_util.find_variables(aggregate_element.tuple) # "Aggregate head" vars
            if self.unsafe_vars is not None:
                ignore_vars = ignore_vars.union(self.unsafe_vars)
            aggregate_condition, tau_rules = self.base_transformer.projection(aggregate_element.condition, self.tau, context_rule=self.rule, ignore_vars=ignore_vars)
            aggregate_element.condition = aggregate_condition
            for tau_rule in tau_rules:
                self.base_transformer.add_stm(tau_rule)
        return body_aggregate
